////////////////////////////////////////////////////////////////////////////////////////////////////
/// @file	AVR_Device_Phenix.c
/// @author	Simon Qian(webmaster@SimonQian.com)
/// @date	Last Update on 2008-04-07
/// @brief	Offline programming part of AVRminiProg
/// @par	license
/// 		AVRminiProg is distributed under the terms and conditions of GPLv2, see License.txt
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <string.h>

#include "../../type.h"
#include "../../app_cfg.h"

#include "../../Delay/delay.h"

#include "../../MCU/gpio.h"

#include "../AVR_Device_Const.h"
#include "../AVR_Device_Common.h"
#include "AVR_Device_Phenix.h"

#if AVR_DEVICE_OFFLINE_EN

#include "AVR_Device_Phenix_ProgramBlock.h"

/// Package Info Struct
typedef struct tag_PackageInfo
{
	uint8 addr;
	uint8 len;
}PackageInfo;

/// Is the Programming Data Valid?
uint8 AVRP_Phenix_ValidByte;
/// Number of Interfaces
uint8 AVRP_Phenix_NumOfInterfaces;
/// Number of Targets
uint8 AVRP_Phenix_NumOfTargets;

/// Current Index of target
uint8 AVRP_Phenix_TargetIdx;
/// Interfaces supported by Current Target
uint8 AVRP_Phenix_InterfaceMsk;
/// Address of Current Target
AVRP_Phenix_DataAddrType AVRP_Phenix_TargetAddr;

/// Read External Configure to detimine which target to program
/// @return
void AVRP_Phenix_ReadConfig(void)
{
	AVRP_Phenix_TargetIdx = 0;
}

/// Read System Informatin from Programming Data
/// @return		Success or Not
uint8 AVRP_Phenix_ReadSystemInfo(void)
{
	uint8 ret = AVRP_Phenix_Error_Success;

	AVRP_Phenix_ValidByte = AVRP_Phenix_GetDataByte(AVRP_Phenix_SysAddr_Valid);
	AVRP_Phenix_NumOfInterfaces = AVRP_Phenix_GetDataByte(AVRP_Phenix_SysAddr_NumOfInterfaces);
	AVRP_Phenix_NumOfTargets = AVRP_Phenix_GetDataByte(AVRP_Phenix_SysAddr_NumOfTargets);

	// Programming Data Exists
	if(AVRP_Phenix_ValidByte || (AVRP_Phenix_NumOfInterfaces > AVRP_Phenix_Constraints_MaxNumOfInterfaces) || (AVRP_Phenix_NumOfTargets > AVRP_Phenix_Constraints_MaxNumOfTargets))
	{
		ret = AVRP_Phenix_Error_ParameterError;
		goto Check_End;
	}
	else
	{
		

		goto Check_End;
	}

Check_End:
	return ret;
}

/// Read Target Information
/// @param[in]	iTarget	Index of Target
/// @return		Success or Not
uint8 AVRP_Phenix_ReadTargetInfo(uint8 iTarget)
{
	uint8 ret = AVRP_Phenix_Error_Success;

	AVRP_Phenix_TargetAddr = AVRP_Phenix_GetAddr(AVRP_Phenix_SysAddr_TargetDataArray + iTarget * AVRP_Phenix_SysInfo_AddrArraySize);
	AVRP_Phenix_InterfaceMsk = AVRP_Phenix_GetDataByte(AVRP_Phenix_TargetAddr + AVRP_Phenix_TargetAddr_Interfaces);

	return ret;
}

/// Call Programmer Implemented by AVRminiProg to do the programming
/// @param[in]	addrInterface		Address of Current Interface in Flash
/// @param[in]	addrProgrammingData	Address of Current Programming Data in Flash
/// @param[in]	pProgrammingData	Address of Programming Data in RAM, if not 0, use this to replace addrProgrammingData
/// @param[in]	iCmd				Call Command
/// @param[in]	addr				Address to be set in Command
/// @return		Success or Not
uint8 AVRP_Phenix_CallProgrammer(AVRP_Phenix_DataAddrType addrInterface,AVRP_Phenix_DataAddrType addrProgrammingData,uint8* pProgrammingData,uint8 iCmd,uint8* addr)
{
	uint8 ret = AVRP_Phenix_Error_Success;
	uint8 dat_idx,addr_idx,addr_len;
	uint16 dat_len;
	AVRP_Phenix_DataAddrType pkgAddr;
	PackageInfo tmpPackageInfo;

	AVRP_Phenix_GetPackageInfo(&tmpPackageInfo,addrInterface,iCmd);

	iCmd |= AVRP_Phenix_GetDataByte(addrInterface + AVRP_Phenix_InterfaceAddr_ProgramMode);
	AVR_Device_DataBuff[0] = iCmd;

	pkgAddr = addrInterface + AVRP_Phenix_InterfaceAddr_PackageAddr + tmpPackageInfo.addr;
	addr_idx = AVRP_Phenix_GetDataByte(pkgAddr + AVRP_Phenix_PackageAddr_AddrIdx);
	addr_len = AVRP_Phenix_GetDataByte(pkgAddr + AVRP_Phenix_PackageAddr_AddrLen);
	dat_idx = AVRP_Phenix_GetDataByte(pkgAddr + AVRP_Phenix_PackageAddr_DataIdx);
	dat_len = AVRP_Phenix_GetDataWord(pkgAddr + AVRP_Phenix_PackageAddr_DataLen);

	AVRP_Phenix_CopyData(AVR_Device_DataBuff + 1,pkgAddr + AVRP_Phenix_PackageAddr_PackageStart,dat_idx);
	if(dat_len)
	{
		if(pProgrammingData)
			memcpy(AVR_Device_DataBuff + 1 + dat_idx,pProgrammingData,dat_len);
		else
			AVRP_Phenix_CopyData(AVR_Device_DataBuff + 1 + dat_idx,addrProgrammingData,dat_len);
		AVRP_Phenix_CopyData(AVR_Device_DataBuff + 1 + dat_idx + dat_len,pkgAddr + AVRP_Phenix_PackageAddr_PackageStart + dat_idx,tmpPackageInfo.len - dat_idx);
	}
	if(addr_len)
	{
		if(addr_len >= AVRP_Phenix_AddrType_SpecialStart)
		{
			// special operation
			switch(addr_len)
			{
#if AVRP_ISP_SHOULD_ENABLE
			case AVRP_Phenix_AddrType_ISPWriteFuse:
				switch(addr[0])
				{
				case 0:
					break;
				case 1:
					AVR_Device_DataBuff[2] |= 0x08;
					break;
				case 2:
					AVR_Device_DataBuff[2] |= 0x04;
					break;
				default:
					return AVRP_Phenix_Error_ParameterError;
				}
				break;
			case AVRP_Phenix_AddrType_ISPReadFuse:
				switch(addr[0])
				{
				case 0:
					break;
				case 1:
					AVR_Device_DataBuff[2] |= 0x08;
					AVR_Device_DataBuff[3] |= 0x08;
					break;
				case 2:
					AVR_Device_DataBuff[3] |= 0x08;
					break;
				default:
					return AVRP_Phenix_Error_ParameterError;
				}
				break;
#endif
			default:
				return AVRP_Phenix_Error_ParameterError;
			}
		}
		else if(addr != (uint8*)0)
			memcpy(AVR_Device_DataBuff + 1 + addr_idx,addr,addr_len);
	}

	dat_len = tmpPackageInfo.len + dat_len;
#if AVRP_Phenix_Debug
	AVRP_Phenix_Debug_Out(AVR_Device_DataBuff,dat_len + 1);
	AVR_Device_DataBuff[1] = 0;
#else
	if(dat_len > 0)
		AVRP_ProcessCmd(AVR_Device_DataBuff,dat_len + 1);
#endif
	if(AVR_Device_DataBuff[1])
		ret = 2 + iCmd;

	return ret;
}

/// Perform Programming Operation
/// @param[in]	iInterface	Index of Current Interface
/// @return		Success or Not
uint8 AVRP_Phenix_DoProgramming(uint8 iInterface)
{
	uint8 ret = 0;
	uint16 operation;
	AVRP_Phenix_DataAddrType addrInterface,addrProgrammingData;
	uint8 addr;
	uint16 iLenTmp = 0;		// avoid warning
	uint16 NumOfPages,i,j;
	uint8 EEPageSize;
	uint16 FlashPageSize;
	uint8 tmp8,tmpData;

	if(!((1 << iInterface) & AVRP_Phenix_InterfaceMsk))
		return AVRP_Phenix_Error_ParameterError;

	addrInterface = AVRP_Phenix_GetAddr(AVRP_Phenix_SysAddr_InterfaceDataArray + iInterface * AVRP_Phenix_SysInfo_AddrArraySize);
	EEPageSize = AVRP_Phenix_GetDataByte(addrInterface + AVRP_Phenix_InterfaceAddr_SizeOfEEPage);
	FlashPageSize = AVRP_Phenix_GetDataByte(addrInterface + AVRP_Phenix_InterfaceAddr_SizeOfFlashPage);

	operation = AVRP_Phenix_GetDataWord(AVRP_Phenix_TargetAddr + AVRP_Phenix_TargetAddr_Operations);

#if AVRP_HVPP_SHOULD_ENABLE || AVRP_HVSP_SHOULD_ENABLE
	// Set PP_Stack, do nothing if not HV
	ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_SPECIAL_OPERATION,(uint8*)0);
	if(ret)
		goto Program_End;
#endif

#if AVRP_ISP_SHOULD_ENABLE
	// Set Programming Speed
	addr = AVRP_Phenix_GetDataByte(AVRP_Phenix_TargetAddr + AVRP_Phenix_TargetAddr_Speed);
	if(addr & 0x80)
	{
		AVR_Device_DataBuff[0] = AVRP_CMD_SET_PARAMETER;
		AVR_Device_DataBuff[1] = AVRP_PARAM_SCK_DURATION;
		AVR_Device_DataBuff[2] = addr & 0x7F;
		AVRP_ProcessCmd(AVR_Device_DataBuff,3);
	}
#endif

	// Enter Program Mode
	ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_ENTER_PROGMODE,(uint8*)0);
	if(ret)
		goto Program_End;

#if AVRP_Phenix_OperationEn_CheckSignature
	// Check Signature
	if(!(operation & (1 << AVRP_Phenix_Operation_CheckSignature)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Signature);

		for(addr = 0;addr < 3;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_SIGNATURE,(uint8*)&addr);
			if(ret)
				goto Program_End;
			if(AVR_Device_DataBuff[2] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
			{
				ret = 2 + AVRP_Phenix_Operation_CheckSignature;
				goto Program_End;
			}
		}		
	}
#endif

#if AVRP_Phenix_OperationEn_ProgramFuseBefore
	// Program Fuse Before
	if(!(operation & (1 << AVRP_Phenix_Operation_ProgramFuseBefore)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_FuseBefore);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		// Program
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,addrProgrammingData++,(uint8*)0,AVRP_CMD_PROGRAM_FUSE,(uint8*)&addr);
			if(ret)
				goto Program_End;
		}
		// Verify
		addrProgrammingData -= iLenTmp;
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_FUSE,(uint8*)&addr);
			if(ret)
				goto Program_End;
			if(AVR_Device_DataBuff[2] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
			{
				ret = 2 + AVRP_Phenix_Operation_ProgramFuseBefore;
				goto Program_End;
			}
		}
	}
#endif

#if AVRP_Phenix_OperationEn_EraseChip
	// Chip Erase
	if(!(operation & (1 << AVRP_Phenix_Operation_EraseChip)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_EraseCmd);

		addr = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_CHIP_ERASE,&addr);
		if(ret)
			goto Program_End;
	}
#endif

#if AVRP_Phenix_OperationEn_ProgramFlash
	// Program Flash
	if(!(operation & (1 << AVRP_Phenix_Operation_ProgramFlash)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Flash);

		NumOfPages = AVRP_Phenix_GetDataWord(addrProgrammingData);
		addrProgrammingData += 2;

		// Set Address to 0
		AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
		AVR_Device_DataBuff[1] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[2] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[3] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[4] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVRP_ProcessCmd(AVR_Device_DataBuff,5);

		// 2 bytes PageSize, not used
		addrProgrammingData += 2;

		for(i = 0;i < NumOfPages;i++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,addrProgrammingData,(uint8*)0,AVRP_CMD_PROGRAM_FLASH,(uint8*)0);
			addrProgrammingData += FlashPageSize;
			if(ret)
				goto Program_End;
		}
	}
#endif

#if AVRP_Phenix_OperationEn_VerifyFlash
	// Verify Flash
	if(!(operation & (1 << AVRP_Phenix_Operation_VerifyFlash)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Flash);

		NumOfPages = AVRP_Phenix_GetDataWord(addrProgrammingData);
		addrProgrammingData += 2;

		// Set Address to 0
		AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
		AVR_Device_DataBuff[1] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[2] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[3] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[4] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVRP_ProcessCmd(AVR_Device_DataBuff,5);

		// 2 bytes PageSize, not used
		addrProgrammingData += 2;

		for(i = 0;i < NumOfPages;i++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_FLASH,(uint8*)0);
			if(ret)
				goto Program_End;
			for(j = 0;j < FlashPageSize;j++)
			{
				if(AVR_Device_DataBuff[2 + j] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
				{
					ret = 2 + AVRP_Phenix_Operation_VerifyFlash;
					goto Program_End;
				}
			}
		}
	}
#endif

#if AVRP_Phenix_OperationEn_ProgramEEPROM
	// Program EEPROM
	if(!(operation & (1 << AVRP_Phenix_Operation_ProgramEEPROM)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_EEPROM);

		NumOfPages = AVRP_Phenix_GetDataWord(addrProgrammingData);
		addrProgrammingData += 2;

		// Set Address to 0
		AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
		AVR_Device_DataBuff[1] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[2] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[3] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[4] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVRP_ProcessCmd(AVR_Device_DataBuff,5);

		// 2 bytes PageSize, not used
		addrProgrammingData += 2;

		for(i = 0;i < NumOfPages;i++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,addrProgrammingData,(uint8*)0,AVRP_CMD_PROGRAM_EEPROM,(uint8*)0);
			addrProgrammingData += EEPageSize;
			if(ret)
				goto Program_End;
		}
	}
#endif

#if AVRP_Phenix_OperationEn_VerifyEEPROM
	// Verify EEPROM
	if(!(operation & (1 << AVRP_Phenix_Operation_VerifyEEPROM)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_EEPROM);

		NumOfPages = AVRP_Phenix_GetDataWord(addrProgrammingData);
		addrProgrammingData += 2;

		// Set Address to 0
		AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
		AVR_Device_DataBuff[1] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[2] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[3] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVR_Device_DataBuff[4] = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		AVRP_ProcessCmd(AVR_Device_DataBuff,5);

		// 2 bytes PageSize, not used
		addrProgrammingData += 2;

		for(i = 0;i < NumOfPages;i++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_EEPROM,(uint8*)0);
			if(ret)
				goto Program_End;
			for(j = 0;j < EEPageSize;j++)
			{
				if(AVR_Device_DataBuff[2 + j] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
				{
					ret = 2 + AVRP_Phenix_Operation_VerifyEEPROM;
					goto Program_End;
				}
			}
		}
	}
#endif

#if AVRP_Phenix_OperationEn_ProgramFuseAfter
	// Program Fuse After
	if(!(operation & (1 << AVRP_Phenix_Operation_ProgramFuseAfter)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_FuseAfter);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		// Program
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,addrProgrammingData++,(uint8*)0,AVRP_CMD_PROGRAM_FUSE,(uint8*)&addr);
			if(ret)
				goto Program_End;
		}
	}
#endif

#if AVRP_Phenix_OperationEn_VerifyFuse
	// Verify Fuse
	if(!(operation & (1 << AVRP_Phenix_Operation_VerifyFuse)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_FuseAfter);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_FUSE,(uint8*)&addr);
			if(ret)
				goto Program_End;
			if(AVR_Device_DataBuff[2] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
			{
				ret = 2 + AVRP_Phenix_Operation_VerifyFuse;
				goto Program_End;
			}
		}
	}
#endif

#if AVRP_Phenix_OperationEn_CalibrateOSC
	// Do Calibrate
	if(!(operation & (1 << AVRP_Phenix_Operation_CalibrateOSC)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Calibrate);

		ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_OSCCAL,(uint8*)0);
		if(ret)
			goto Program_End;
	}
#endif

#if AVRP_Phenix_OperationEn_WriteOSCCAL
	// Write OSC
	if(!(operation & (1 << AVRP_Phenix_Operation_WriteOSCCAL)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_OSCCAL);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		for(i = 0;i < iLenTmp;i++)
		{
			tmp8 = AVRP_Phenix_GetDataByte(addrProgrammingData++);

			// Read Data
			addr = tmp8 & 0x03;
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_OSCCAL,(uint8*)&addr);
			if(ret)
				goto Program_End;

			if(tmp8 & 0x80)
				j = FlashPageSize;
			else
				j = EEPageSize;

			tmpData = AVR_Device_DataBuff[2];
			NumOfPages = AVRP_Phenix_GetDataWord(addrProgrammingData);
			addrProgrammingData += 2;

			addr = NumOfPages % FlashPageSize;
			NumOfPages -= addr;

			AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
			AVR_Device_DataBuff[1] = 0;
			AVR_Device_DataBuff[2] = 0;
			AVR_Device_DataBuff[3] = NumOfPages >> 9;
			AVR_Device_DataBuff[4] = NumOfPages >> 1;
			AVRP_ProcessCmd(AVR_Device_DataBuff,5);

			for(i =  CMD_MAX_PKGSIZE - j;i < CMD_MAX_PKGSIZE;i++)
				AVR_Device_DataBuff[i] = 0xFF;
			AVR_Device_DataBuff[CMD_MAX_PKGSIZE - j + addr] = tmpData;

			if(tmp8 & 0x80)
				ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)&AVR_Device_DataBuff[CMD_MAX_PKGSIZE - j],AVRP_CMD_PROGRAM_FLASH,(uint8*)0);
			else
				ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)&AVR_Device_DataBuff[CMD_MAX_PKGSIZE - j],AVRP_CMD_PROGRAM_EEPROM,(uint8*)0);
			if(ret)
				goto Program_End;

			AVR_Device_DataBuff[0] = AVRP_CMD_LOAD_ADDRESS;
			AVR_Device_DataBuff[1] = 0;
			AVR_Device_DataBuff[2] = 0;
			AVR_Device_DataBuff[3] = NumOfPages >> 9;
			AVR_Device_DataBuff[4] = NumOfPages >> 1;
			AVRP_ProcessCmd(AVR_Device_DataBuff,5);

			// Verify
			if(tmp8 & 0x80)
				ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_FLASH,(uint8*)0);
			else
				ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_EEPROM,(uint8*)0);
			if(ret)
				goto Program_End;

			if(AVR_Device_DataBuff[2 + addr)] != tmpData)
			{
				ret = AVRP_Phenix_Operation_WriteOSCCAL;
				goto Program_End;
			}
		}
	}
#endif

#if AVRP_Phenix_OperationEn_ProgramLockbits
	// Program Lockbits
	if(!(operation & (1 << AVRP_Phenix_Operation_ProgramLockbits)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Lockbits);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		// Program
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,addrProgrammingData++,(uint8*)0,AVRP_CMD_PROGRAM_LOCKBIT,(uint8*)&addr);
			if(ret)
				goto Program_End;
		}
	}
#endif

#if AVRP_Phenix_OperationEn_VerifyLockbits
	// Verify Lickbits
	if(!(operation & (1 << AVRP_Phenix_Operation_VerifyLockbits)))
	{
		addrProgrammingData = AVRP_Phenix_GetTargetAddr(AVRP_Phenix_TargetAddr,AVRP_Phenix_ProgramDataArrayIdx_Lockbits);

		iLenTmp = AVRP_Phenix_GetDataByte(addrProgrammingData++);
		for(addr = 0;addr < iLenTmp;addr++)
		{
			ret = AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_READ_LOCKBIT,(uint8*)&addr);
			if(ret)
				goto Program_End;
			if(AVR_Device_DataBuff[2] != AVRP_Phenix_GetDataByte(addrProgrammingData++))
			{
				ret = 2 + AVRP_Phenix_Operation_VerifyLockbits;
				goto Program_End;
			}
		}
	}
#endif

Program_End:
	// Leave Program Mode
	AVRP_Phenix_CallProgrammer(addrInterface,0,(uint8*)0,AVRP_CMD_EXIT_PROGMODE,(uint8*)0);

	return ret;
}

/// Offline Programmer Initialization
/// @return		Success or Not
uint8 AVRP_Phenix_Init(void)
{
	uint8 ret = 0;

	ret = AVRP_Phenix_ReadSystemInfo();
	ret += AVRP_Phenix_ReadTargetInfo(AVRP_Phenix_TargetIdx);

	return ret;
}

/// Main function of Offline Programmer
/// @return
void AVRP_Phenix_Main(void)
{
	uint8 mskInterface,iInterface = 0;

	AVRP_Phenix_ReadConfig();

	if(AVRP_Phenix_Init())
		goto OnFatalError;

	while(PROGKEY_IsDown());
	while(1)
	{
Wait_Key_Down:
 		while(!PROGKEY_IsDown());
		DelayMS(0xFF);
		if(!PROGKEY_IsDown())
			goto Wait_Key_Down;

		LED_Off(LED_Green | LED_Red);
		// Do Programming
		mskInterface = 0;
		while(AVRP_Phenix_DoProgramming(iInterface))
		{
			mskInterface |= (1 << iInterface);
			if((mskInterface & AVRP_Phenix_InterfaceMsk) == AVRP_Phenix_InterfaceMsk)
				break;
			if(++iInterface > AVRP_Phenix_Constraints_MaxNumOfInterfaces)
				iInterface = 0;
		}

Wait_Key_Up:
		while(PROGKEY_IsDown());
		DelayMS(0xFF);
		if(PROGKEY_IsDown())
			goto Wait_Key_Up;

		if((mskInterface & AVRP_Phenix_InterfaceMsk) == AVRP_Phenix_InterfaceMsk)
			LED_On(LED_Red);
		else
			LED_On(LED_Green);
	}

OnFatalError:
	while(1)
	{
		uint8 dly;

		if(LED_IsOn(LED_Green | LED_Red))
			LED_Off(LED_Green | LED_Red);
		else
			LED_On(LED_Green | LED_Red);
		for(dly = 1;dly;dly++)
			DelayMS(0xFF);
	}
}

#endif
